use std::cmp::min;

use image::{DynamicImage, GenericImageView, Rgba, RgbaImage};
use image::imageops::{FilterType, overlay};
use imageproc::drawing::{Blend, Canvas, draw_filled_circle_mut};

use crate::config::image_processors::ProcessorParams;
use crate::img_processors::ImageProcessor;
use crate::result::{MBGError, MBGResult};

pub struct Border {
    pub params: ProcessorParams,
    pub circle: bool,
    pub color: String,
    pub width: u32,
}


impl Border {
    fn crop_circle(&self, img: &DynamicImage) -> MBGResult<DynamicImage> {
        let img_height = image::GenericImageView::height(img);
        let img_width = image::GenericImageView::width(img);
        let transparent_pixel = Rgba([0u8, 0u8, 0u8, 0u8]);
        let mut mask = Blend(RgbaImage::from_pixel(img_width * 3, img_height * 3, transparent_pixel));
        let center = ((mask.width() / 2) as i32, (mask.height() / 2) as i32);
        let radius = min(center.0, center.1);
        let black_pixel = Rgba([0u8, 0u8, 0u8, 255u8]);
        draw_filled_circle_mut(&mut mask, center, radius, black_pixel);
        let mut mask = DynamicImage::ImageRgba8(mask.0);
        mask = mask.resize(img_width, img_height, FilterType::Nearest);
        let mut res = mask.clone();
        for (x, y, pix) in mask.pixels() {
            let pixel = image::GenericImageView::get_pixel(img, x, y).0;
            res.draw_pixel(x, y, Rgba([pixel[0], pixel[1], pixel[2], pix.0[3]]));
        }
        Ok(res)
    }

    fn draw_border(&self, img: &DynamicImage) -> MBGResult<DynamicImage> {
        let img_width = image::GenericImageView::width(img);
        let img_height = image::GenericImageView::height(img);
        let center = (((img_width / 2) + self.width), ((img_height / 2) + self.width));
        let target_color = read_color::rgb_maybe_a(&mut self.color.chars());
        if target_color.is_none() {
            return Err(
                MBGError::ConfigErr(String::from("Can't parse border color. Please use hex value."))
            );
        }
        let target_color = target_color.unwrap();
        let target_color = match target_color {
            ([r, g, b, ], Some(a)) => Rgba([r, g, b, a]),
            ([r, g, b, ], None) => Rgba([r, g, b, 255u8]),
        };
        if self.circle {
            let mut mask = DynamicImage::ImageRgba8(RgbaImage::from_pixel(
                img_width + self.width * 2,
                img_height + self.width * 2,
                Rgba([0u8, 0u8, 0u8, 0u8]),
            ));
            let radius = min(center.0, center.1);
            let cir_center = (center.0 as i32, center.1 as i32);
            draw_filled_circle_mut(&mut mask, cir_center, radius as i32, target_color);
            overlay(&mut mask, img, self.width, self.width);
            Ok(mask)
        } else {
            let mut mask = DynamicImage::ImageRgba8(RgbaImage::from_pixel(
                img_width + self.width * 2,
                img_height + self.width * 2,
                target_color,
            ));
            overlay(&mut mask, img, self.width, self.width);
            Ok(mask)
        }
    }
}

impl ImageProcessor for Border {
    fn process(&self, img: &DynamicImage) -> MBGResult<DynamicImage> {
        let mut res = img.clone();
        if self.circle {
            res = self.crop_circle(&res)?;
        }
        if self.width > 0 {
            res = self.draw_border(&res)?;
        }
        Ok(res)
    }
}
