use image::imageops::overlay;
use image::{DynamicImage, GenericImageView};

use crate::config::image_processors::ProcessorParams;
use crate::config::Config;
use crate::display::get_max_resolution;
use crate::result::{MBGError, MBGResult};

pub mod blur;
pub mod border;
pub mod crop;
pub mod none;
pub mod scale;

#[derive(Serialize, Clone, Deserialize, Debug)]
pub enum Processors {
    None,
    Blur,
    Scale,
    Crop {
        #[serde(default)]
        width: u32,
        #[serde(default)]
        height: u32,
        #[serde(default)]
        fit: bool,
    },
    Border {
        #[serde(default)]
        circle: bool,
        #[serde(default)]
        color: String,
        #[serde(default)]
        width: u32,
    },
}

pub trait ImageProcessor {
    fn process(&self, img: &DynamicImage) -> MBGResult<DynamicImage>;
}

pub trait GetLayer {
    fn get_layer(&self) -> usize;
}

pub trait ImgFilterTrait: GetLayer + ImageProcessor {}

macro_rules! impl_params_getter {
    ($($inst:path),*) => {
        $(impl GetLayer for $inst {
            fn get_layer(&self) -> usize {
                self.params.layer
            }
        }
        impl ImgFilterTrait for $inst{}
        )*
    };
}

impl Default for Processors {
    fn default() -> Self {
        Self::None
    }
}

pub fn process_image(img_bytes: Vec<u8>) -> MBGResult<DynamicImage> {
    let conf = Config::get_props();
    let (screen_height, screen_width) = get_max_resolution()?;
    debug!("Maximum resolution: {}x{}", screen_width, screen_height);
    let image = image::load_from_memory(img_bytes.as_slice())?;
    let mut layers = vec![image.clone()];
    for params in conf.processors.into_iter() {
        let processor = mapper(params, screen_height, screen_width);
        if let Some(layer_img) = layers.get(processor.get_layer()) {
            layers[processor.get_layer()] = processor.process(layer_img)?;
        } else {
            layers.push(processor.process(&image.clone())?);
        }
        debug!("Filter applied");
    }
    if layers.is_empty() {
        return Err(MBGError::ImageError(String::from(
            "No processors was found.",
        )));
    }
    if conf.blender.is_empty() {
        Ok(layers[0].clone())
    } else {
        blend_layers(layers, conf.blender)
    }
}

pub fn blend_layers(layers: Vec<DynamicImage>, blender: Vec<u8>) -> MBGResult<DynamicImage> {
    let mut res = layers[0].clone();
    for index in blender.into_iter() {
        if let Some(layer) = layers.get(usize::from(index)) {
            let pos_x = (res.width() / 2) - (layer.width() / 2);
            let pos_y = (res.height() / 2) - (layer.height() / 2);
            overlay(&mut res, layer, pos_x, pos_y);
        }
    }
    Ok(res)
}

impl_params_getter!(
    none::NoneProcessor,
    blur::BlurProcessor,
    scale::ScaleProcessor,
    crop::CropProcessor,
    border::Border
);

fn mapper(
    params: ProcessorParams,
    screen_height: i32,
    screen_width: i32,
) -> Box<dyn ImgFilterTrait> {
    match params.clone().effect {
        Processors::None => Box::new(none::NoneProcessor { params }),
        Processors::Blur => Box::new(blur::BlurProcessor { params }),
        Processors::Scale => Box::new(scale::ScaleProcessor {
            params,
            screen_height,
            screen_width,
        }),
        Processors::Crop { width, height, fit } => Box::new(crop::CropProcessor {
            params,
            screen_height,
            screen_width,
            height,
            width,
            fit,
        }),
        Processors::Border {
            circle,
            color,
            width,
        } => Box::new(border::Border {
            params,
            circle,
            color,
            width,
        }),
    }
}